@@ -14,7 +14,13 @@ from dbnd._core.errors import DatabandError, DatabandRuntimeError, friendly_erro
 from dbnd._core.log.logging_utils import override_log_formatting
 from dbnd._core.task_run.task_run_error import TaskRunError
 from dbnd._core.utils.timezone import utcnow
-from dbnd_airflow.airflow_extensions.dal import get_airflow_task_instance_state
+from dbnd_airflow.airflow_extensions.dal import (
+    get_airflow_task_instance,
+    get_airflow_task_instance_state,
+)
+from dbnd_airflow_contrib.airflow_task_instance_retry_controller import (
+    AirflowTaskInstanceRetryController,
+)
 from dbnd_docker.kubernetes.kube_resources_checker import DbndKubeResourcesChecker
 from dbnd_docker.kubernetes.kubernetes_engine_config import (
     KubernetesEngineConfig,
@@ -187,12 +193,11 @@ class DbndKubernetesClient(object):
         except DatabandError as ex:
             error = TaskRunError.buid_from_ex(ex, task_run)
 
-        task_state = get_airflow_task_instance_state(task_run=task_run)
-
-        logger.debug("task airflow state: %s ", task_state)
+        airflow_task_state = get_airflow_task_instance_state(task_run=task_run)
+        logger.debug("task airflow state: %s ", airflow_task_state)
         from airflow.utils.state import State
 
-        if task_state == State.FAILED:
+        if airflow_task_state == State.FAILED:
             # let just notify the error, so we can show it in summary it
             # we will not send it to databand tracking store
             task_run.set_task_run_state(TaskRunState.FAILED, track=False, error=error)
@@ -206,7 +211,29 @@ class DbndKubernetesClient(object):
                 ),
             )
         else:
-            task_run.set_task_run_state(TaskRunState.FAILED, track=True, error=error)
+            if airflow_task_state == State.QUEUED:
+                # Special case - no airflow code has been run in the pod at all. Must increment try number and send
+                # to retry if exit code is matching
+                if not pod_ctrl.handle_pod_retry(
+                    pod_data, task_run, self.engine_config, increment_try_number=True
+                ):
+                    # No retry was sent
+                    task_run.set_task_run_state(
+                        TaskRunState.FAILED, track=True, error=error
+                    )
+            elif airflow_task_state == State.RUNNING:
+                # Task was killed unexpectedly -- probably pod failure in K8s - Possible retry attempt
+                if not pod_ctrl.handle_pod_retry(
+                    pod_data, task_run, self.engine_config
+                ):
+                    # No retry was sent
+                    task_run.set_task_run_state(
+                        TaskRunState.FAILED, track=True, error=error
+                    )
+            else:
+                task_run.set_task_run_state(
+                    TaskRunState.FAILED, track=True, error=error
+                )
             if logs:
                 task_run.tracker.save_task_run_log("\n".join(logs))
 
@@ -430,16 +457,16 @@ class DbndPodCtrl(object):
         detach_run = detach_run or kc.detach_run
         if kc.show_pod_log:
             logger.info(
-                "%s is True,  will send every docker in blocking mode",
-                kc.task_name,
+                "%s is True, %s will send every docker in blocking mode",
                 "show_pod_logs",
+                kc.task_name,
             )
             detach_run = False
         if kc.debug:
             logger.info(
-                "%s is True,  will send every docker in blocking mode",
-                kc.task_name,
+                "%s is True, %s will send every docker in blocking mode",
                 "debug",
+                kc.task_name,
             )
             detach_run = False
 
@@ -491,6 +518,100 @@ class DbndPodCtrl(object):
         self.wait()
         return self
 
+    def handle_pod_retry(
+        self, pod_data, task_run, engine_config, increment_try_number=False
+    ):
+        metadata = pod_data.metadata
+        logger.debug(
+            "Reached handle pod retry for pod %s! Looking for pod exit code",
+            metadata.name,
+        )
+        pod_exit_code = self._try_get_pod_exit_code(pod_data)
+        if not pod_exit_code:
+            # Couldn't find an exit code - container is still alive - wait for the next event
+            logger.debug("No exit code found for pod %s, doing nothing", metadata.name)
+            return False
+        logger.info("Found pod exit code %d for pod %s", pod_exit_code, metadata.name)
+        if self._should_pod_be_retried(pod_exit_code, engine_config):
+            retry_count, retry_delay = self._get_pod_retry_parameters(
+                pod_exit_code, engine_config
+            )
+            task_instance = get_airflow_task_instance(task_run)
+            task_instance.max_tries = retry_count
+            """
+            TaskInstance has no retry delay property, it is gathered from the DbndOperator.
+            The operator's values are overridden and taken from configuration if the engine is running K8s.
+            See dbnd_operator.py
+            """
+            return self._schedule_pod_for_retry(
+                metadata,
+                retry_count,
+                retry_delay,
+                task_instance,
+                task_run,
+                increment_try_number,
+            )
+        else:
+            logger.debug(
+                "Pod %s was not scheduled for retry because its exit code %d was not found in config",
+                metadata.name,
+                pod_exit_code,
+            )
+            return False
+
+    @staticmethod
+    def _get_pod_retry_parameters(pod_exit_code, engine_config):
+        # Configuration stores keys of dictionary as strings, hence we call str() on the pod exit code
+        retry_count = engine_config.pod_exit_code_to_retry_count[str(pod_exit_code)]
+        retry_delay = engine_config.pod_retry_delay
+        return retry_count, retry_delay
+
+    @staticmethod
+    def _should_pod_be_retried(pod_exit_code, engine_config):
+        # Configuration stores keys of dictionary as strings, hence we call str() on the pod exit code
+        return str(pod_exit_code) in engine_config.pod_exit_code_to_retry_count
+
+    @staticmethod
+    def _schedule_pod_for_retry(
+        metadata,
+        retry_count,
+        retry_delay,
+        task_instance,
+        task_run,
+        increment_try_number,
+    ):
+
+        task_instance_retry_controller = AirflowTaskInstanceRetryController(
+            task_instance, task_run
+        )
+        if task_instance_retry_controller.schedule_task_instance_for_retry(
+            retry_count, retry_delay, increment_try_number
+        ):
+            logger.info(
+                "Scheduling the pod %s for %s retries with delay of %s",
+                metadata.name,
+                retry_count,
+                retry_delay,
+            )
+            return True
+        else:
+            logger.warning(
+                "Pod %s was not scheduled for retry because it reached the maximum retry limit"
+            )
+            return False
+
+    @staticmethod
+    def _try_get_pod_exit_code(pod_data):
+        found_exit_code = False
+        if pod_data.status.container_statuses:
+            for container_status in pod_data.status.container_statuses:
+                # Searching for the container that was terminated
+                if container_status.state.terminated:
+                    pod_exit_code = container_status.state.terminated.exit_code
+                    found_exit_code = True
+            if found_exit_code:
+                return pod_exit_code
+
 
 def _get_status_log_safe(pod_data):
     try:
